@@ -26,7 +26,10 @@ from botocore.awsrequest import AWSResponse
 from botocore.exceptions import InvalidExpressionError, ConfigNotFound
 from botocore.exceptions import ClientError, ConnectionClosedError
 from botocore.exceptions import InvalidDNSNameError, MetadataRetrievalError
+from botocore.exceptions import ReadTimeoutError
+from botocore.exceptions import ConnectTimeoutError
 from botocore.model import ServiceModel
+from botocore.utils import ensure_boolean
 from botocore.utils import is_json_value_header
 from botocore.utils import remove_dot_segments
 from botocore.utils import normalize_url_path
@@ -54,13 +57,31 @@ from botocore.utils import switch_host_s3_accelerate
 from botocore.utils import deep_merge
 from botocore.utils import S3RegionRedirector
 from botocore.utils import ContainerMetadataFetcher
-from botocore.utils import IMDSFetcher
 from botocore.utils import InstanceMetadataFetcher
+from botocore.utils import IMDSFetcher
+from botocore.utils import BadIMDSRequestError
 from botocore.model import DenormalizedStructureBuilder
 from botocore.model import ShapeResolver
 from botocore.config import Config
 
 
+class TestEnsureBoolean(unittest.TestCase):
+    def test_boolean_true(self):
+        self.assertEqual(ensure_boolean(True), True)
+
+    def test_boolean_false(self):
+        self.assertEqual(ensure_boolean(False), False)
+
+    def test_string_true(self):
+        self.assertEqual(ensure_boolean('True'), True)
+
+    def test_string_false(self):
+        self.assertEqual(ensure_boolean('False'), False)
+
+    def test_string_lowercase_true(self):
+        self.assertEqual(ensure_boolean('true'), True)
+
+
 class TestIsJSONValueHeader(unittest.TestCase):
     def test_no_serialization_section(self):
         shape = mock.Mock()
@@ -1703,8 +1724,9 @@ class TestContainerMetadataFetcher(unittest.TestCase):
             self.fake_response(status_code=200, body=b'Not JSON'),
         )
         fetcher = self.create_fetcher()
-        with self.assertRaises(MetadataRetrievalError):
+        with self.assertRaises(MetadataRetrievalError) as e:
             fetcher.retrieve_uri('/foo?id=1')
+        self.assertNotIn('Not JSON', str(e.exception))
         # Should have tried up to RETRY_ATTEMPTS.
         self.assertEqual(self.http.send.call_count, fetcher.RETRY_ATTEMPTS)
 
@@ -1765,6 +1787,13 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
             'Token': 'spam-token',
             'Expiration': 'something',
         }
+        self._expected_creds = {
+            'access_key': self._creds['AccessKeyId'],
+            'secret_key': self._creds['SecretAccessKey'],
+            'token': self._creds['Token'],
+            'expiry_time': self._creds['Expiration'],
+            'role_name': self._role_name
+        }
 
     def tearDown(self):
         self._urllib3_patch.stop()
@@ -1788,6 +1817,13 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
             creds = self._creds
         self.add_imds_response(body=json.dumps(creds).encode('utf-8'))
 
+    def add_get_token_imds_response(self, token, status_code=200):
+        self.add_imds_response(body=token.encode('utf-8'),
+                               status_code=status_code)
+
+    def add_metadata_token_not_supported_response(self):
+        self.add_imds_response(b'', status_code=404)
+
     def add_imds_connection_error(self, exception):
         self._imds_responses.append(exception)
 
@@ -1815,84 +1851,63 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
         url = 'https://example.com/'
         env = {'AWS_EC2_METADATA_DISABLED': 'false'}
 
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         self.add_get_credentials_imds_response()
 
         fetcher = InstanceMetadataFetcher(base_url=url, env=env)
         result = fetcher.retrieve_iam_role_credentials()
 
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_includes_user_agent_header(self):
         user_agent = 'my-user-agent'
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         self.add_get_credentials_imds_response()
 
         InstanceMetadataFetcher(
             user_agent=user_agent).retrieve_iam_role_credentials()
 
-        headers = self._send.call_args[0][0].headers
-        self.assertEqual(headers['User-Agent'], user_agent)
+        self.assertEqual(self._send.call_count, 3)
+        for call in self._send.calls:
+            self.assertTrue(call[0][0].headers['User-Agent'], user_agent)
 
     def test_non_200_response_for_role_name_is_retried(self):
         # Response for role name that have a non 200 status code should
         # be retried.
+        self.add_get_token_imds_response(token='token')
         self.add_imds_response(
             status_code=429, body=b'{"message": "Slow down"}')
         self.add_get_role_name_imds_response()
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_http_connection_error_for_role_name_is_retried(self):
         # Connection related errors should be retried
+        self.add_get_token_imds_response(token='token')
         self.add_imds_connection_error(ConnectionClosedError(endpoint_url=''))
         self.add_get_role_name_imds_response()
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_empty_response_for_role_name_is_retried(self):
         # Response for role name that have a non 200 status code should
         # be retried.
+        self.add_get_token_imds_response(token='token')
         self.add_imds_response(body=b'')
         self.add_get_role_name_imds_response()
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_non_200_response_is_retried(self):
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         # Response for creds that has a 200 status code but has an empty
         # body should be retried.
@@ -1901,32 +1916,20 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_http_connection_errors_is_retried(self):
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         # Connection related errors should be retried
         self.add_imds_connection_error(ConnectionClosedError(endpoint_url=''))
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_empty_response_is_retried(self):
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         # Response for creds that has a 200 status code but is empty.
         # This should be retried.
@@ -1934,16 +1937,10 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_invalid_json_is_retried(self):
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         # Response for creds that has a 200 status code but is invalid JSON.
         # This should be retried.
@@ -1951,22 +1948,17 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
         self.add_get_credentials_imds_response()
         result = InstanceMetadataFetcher(
             num_attempts=2).retrieve_iam_role_credentials()
-        expected_result = {
-            'access_key': self._creds['AccessKeyId'],
-            'secret_key': self._creds['SecretAccessKey'],
-            'token': self._creds['Token'],
-            'expiry_time': self._creds['Expiration'],
-            'role_name': self._role_name
-        }
-        self.assertEqual(result, expected_result)
+        self.assertEqual(result, self._expected_creds)
 
     def test_exhaust_retries_on_role_name_request(self):
+        self.add_get_token_imds_response(token='token')
         self.add_imds_response(status_code=400, body=b'')
         result = InstanceMetadataFetcher(
             num_attempts=1).retrieve_iam_role_credentials()
         self.assertEqual(result, {})
 
     def test_exhaust_retries_on_credentials_request(self):
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         self.add_imds_response(status_code=400, body=b'')
         result = InstanceMetadataFetcher(
@@ -1974,6 +1966,7 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
         self.assertEqual(result, {})
 
     def test_missing_fields_in_credentials_response(self):
+        self.add_get_token_imds_response(token='token')
         self.add_get_role_name_imds_response()
         # Response for creds that has a 200 status code and a JSON body
         # representing an error. We do not necessarily want to retry this.
@@ -1981,3 +1974,90 @@ class TestInstanceMetadataFetcher(unittest.TestCase):
             body=b'{"Code":"AssumeRoleUnauthorizedAccess","Message":"error"}')
         result = InstanceMetadataFetcher().retrieve_iam_role_credentials()
         self.assertEqual(result, {})
+
+    def test_token_is_included(self):
+        user_agent = 'my-user-agent'
+        self.add_get_token_imds_response(token='token')
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        # Check that subsequent calls after getting the token include the token.
+        self.assertEqual(self._send.call_count, 3)
+        for call in self._send.call_args_list[1:]:
+            self.assertEqual(call[0][0].headers['x-aws-ec2-metadata-token'], 'token')
+        self.assertEqual(result, self._expected_creds)
+
+    def test_metadata_token_not_supported_404(self):
+        user_agent = 'my-user-agent'
+        self.add_imds_response(b'', status_code=404)
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        for call in self._send.call_args_list[1:]:
+            self.assertNotIn('x-aws-ec2-metadata-token', call[0][0].headers)
+        self.assertEqual(result, self._expected_creds)
+
+    def test_metadata_token_not_supported_403(self):
+        user_agent = 'my-user-agent'
+        self.add_imds_response(b'', status_code=403)
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        for call in self._send.call_args_list[1:]:
+            self.assertNotIn('x-aws-ec2-metadata-token', call[0][0].headers)
+        self.assertEqual(result, self._expected_creds)
+
+    def test_metadata_token_not_supported_405(self):
+        user_agent = 'my-user-agent'
+        self.add_imds_response(b'', status_code=405)
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        for call in self._send.call_args_list[1:]:
+            self.assertNotIn('x-aws-ec2-metadata-token', call[0][0].headers)
+        self.assertEqual(result, self._expected_creds)
+
+    def test_metadata_token_not_supported_timeout(self):
+        user_agent = 'my-user-agent'
+        self.add_imds_connection_error(ReadTimeoutError(endpoint_url='url'))
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        for call in self._send.call_args_list[1:]:
+            self.assertNotIn('x-aws-ec2-metadata-token', call[0][0].headers)
+        self.assertEqual(result, self._expected_creds)
+
+    def test_token_not_supported_exhaust_retries(self):
+        user_agent = 'my-user-agent'
+        self.add_imds_connection_error(ConnectTimeoutError(endpoint_url='url'))
+        self.add_get_role_name_imds_response()
+        self.add_get_credentials_imds_response()
+
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+
+        for call in self._send.call_args_list[1:]:
+            self.assertNotIn('x-aws-ec2-metadata-token', call[0][0].headers)
+        self.assertEqual(result, self._expected_creds)
+
+    def test_metadata_token_bad_request_yields_no_credentials(self):
+        user_agent = 'my-user-agent'
+        self.add_imds_response(b'', status_code=400)
+        result = InstanceMetadataFetcher(
+            user_agent=user_agent).retrieve_iam_role_credentials()
+        self.assertEqual(result, {})
